import shapeless._

object JSON3 {
  // Data  
  sealed trait User
  case class Person(names: List[String], age: Int) extends User

  sealed trait JSON
  case class JSValue(r: String) extends JSON {
    override def toString = r
  }
  case class JSArray(items: Seq[JSON]) extends JSON {
    override def toString = items.mkString("[",",","]")
  }
  case class JSObject(items: Map[String,JSON]) extends JSON {
    override def toString = items.map{ case (k,v) => s""""$k":$v""" }.mkString("{",",","}")
  }

  // JSON Conversion
  sealed trait ToJSON[T] {
    def apply(in: T): JSON
  }

  sealed trait Derive[T] {
    def instance[R](
      implicit gen: LabelledGeneric.Aux[T,R],
      toJSON: ToJSON[R]) = ToJSON[T] { t => toJSON(gen.to(t)) }
  }

  object ToJSON {
    def apply[T](f: T => JSON): ToJSON[T] = new ToJSON[T] {
      def apply(in: T) = f(in)
    }

    def derive[T] = new Derive[T] { }  
  }

  def toJSON[T](t: T)(implicit toJSON: ToJSON[T]) = toJSON(t)

  implicit val stringToJSON = 
    ToJSON[String](s => JSValue('"' + s + '"')) // Will break with escapes

  implicit val intToJSON = 
    ToJSON[Int](i => JSValue(i.toString))

  implicit def listToJSON[T](implicit toJSON: ToJSON[T]) = 
    ToJSON[List[T]](l => JSArray(l.map(toJSON(_))))

  implicit def hNilToJSON =
    ToJSON[HNil](_ => JSObject(Map.empty))

  implicit def hConsToJSON[Key <: Symbol, Head, Tail <: HList](
    implicit key: Witness.Aux[Key],
    headJSON: ToJSON[Head],
    tailJSON: ToJSON[Tail]
  ): ToJSON[Head :: Tail] =
    ToJSON { l =>
      tailJSON(l.tail) match {
        case JSObject(fields) => JSObject(fields + (key.value.name -> headJSON(l.head)))
      }
    }

  implicit def personToJSON = ToJSON.derive[Person].instance
}